//
// On Win7 built with MSVC 2008 shellicon refuses to load without this.
// When built with MingW shellicon loads normally.
// Some conflict with treectrl.dll using a different comctl?
//

#ifndef RT_MANIFEST
#define RT_MANIFEST     24
#endif
#ifndef ISOLATIONAWARE_MANIFEST_RESOURCE_ID
#define ISOLATIONAWARE_MANIFEST_RESOURCE_ID 2
#endif
ISOLATIONAWARE_MANIFEST_RESOURCE_ID RT_MANIFEST "shellicon.dll.manifest"

